/****************************************************************
 *                                                              *
 * File          : nvector.c                                    *
 * Last Modified : 17 August 1999                               *
 *--------------------------------------------------------------*
 *                                                              *
 * This is the implementation file for a generic VECTOR         *
 * package. It contains the implementation of the N_Vector      *
 * kernels listed in nvector.h.                                 *
 * Modified for MPI.                                            *
 *                                                              *
 ****************************************************************/

/*......................................................................


                            LEGAL NOTICES


1. NOTICE

This work was performed at the University of California, Lawrence
Livermore National Laboratory (UC LLNL) under contract no.
W-7405-ENG-48 (Contract 48) between the U.S. Department of Energy
(DOE) and The Regents of the University of California (the University)
for the operation of UC LLNL.  Copyright is reserved to the University
for purposes of controlled dissemination, commercialization through
formal licensing, or other disposition under terms of Contract 48; DOE
policies, regulations, and orders; and U.S. statutes.  The rights of
the Federal Government are reserved under Contract 48 subject to the
restrictions agreed upon by the DOE and University as allowed under
DOE Acquisition Letter 97-1.


2. DISCLAIMER

This work was prepared as an account of work sponsored by an agency of
the United States Government.  Neither the United States Government
nor the University of California nor any of their empolyees makes any
warranty, express or implied, or assumes any liability or
responsibility for the accuracy, completeness, or usefulness of any
information, apparatus, product, or process disclosed, or represents
that its use would not infringe privately owned rights.  Reference
herein to any specific commercial products, process, or service by
trade name, trademark, manufacturer, or otherwise, does not
necessarily constitute or imply its endorsement, recommendation, or
favoring by the United States Government or the University of
California.  The views and opinions of authors expressed herein do not
necessarily state or reflect those of the United States Government or
the University of California, and shall not be used for advertising or
product endorsement purposes.


3. LICENSING REQUIREMENTS

Permission is hereby granted to use and copy this software and
documentation for internal non-commercial purposes only, provided that
all UC LLNL identification in the user interface remains unchanged.
Any use, reproduction, modification, or distribution of this softwre
or documentation for commercial purposes requires a license from the
University.  Contact: Lawrence Livermore National Laboratory,
Industrial Partnerships and Commercialization Office, P.O. Box 808,
L-795, Livermore, CA 94551.

......................................................................*/



#include <stdio.h>
#include <stdlib.h>
#include "nvector.h"
#include "llnltyps.h"
#include "llnlmath.h" 


#define ZERO RCONST(0.0)
#define ONE  RCONST(1.0)

/* Error message */

#define BAD_N1  "PVecInitMPI-- Sum of local vector lengths differs from "
#define BAD_N2  "input global length. \n\n"
#define BAD_N    BAD_N1 BAD_N2

/* Private Helper Prototypes */

static real PVecAllReduce(real d, int op, machEnvType machEnv);
/* Reduction operations add/max/min over the processor group */
static void VCopy(N_Vector x, N_Vector z); /* z=x */
static void VSum(N_Vector x, N_Vector y, N_Vector z); /* z=x+y */
static void VDiff(N_Vector x, N_Vector y, N_Vector z); /* z=x-y */
static void VNeg(N_Vector x, N_Vector z); /* z=-x */
/* z=c(x+y) */
static void VScaleSum(real c, N_Vector x, N_Vector y, N_Vector z);
/* z=c(x-y) */
static void VScaleDiff(real c, N_Vector x, N_Vector y, N_Vector z); 
static void VLin1(real a, N_Vector x, N_Vector y, N_Vector z); /* z=ax+y */
static void VLin2(real a, N_Vector x, N_Vector y, N_Vector z); /* z=ax-y */
static void Vaxpy(real a, N_Vector x, N_Vector y); /* y <- ax+y */
static void VScaleBy(real a, N_Vector x); /* x <- ax */


/********************* Exported Functions ************************/


void *PVecInitMPI(MPI_Comm comm,  integer local_vec_length, 
		  integer global_vec_length, int *argc, char ***argv)
{
  int initflag, initerr;
  integer n, Nsum;
  machEnvType env;

  /* Create structure env and begin loading it */
  env = (machEnvType) malloc (sizeof (struct machEnvTypeStruct));
  if (env == NULL) return(NULL);

  env->local_vec_length = local_vec_length;
  env->global_vec_length = global_vec_length;

  MPI_Initialized(&initflag);
  if (!initflag) {
    initerr = MPI_Init(argc,argv);
    if (initerr != MPI_SUCCESS) return(NULL);
    }
  env->init_by_user = initflag;

  env->comm = (comm == NULL ? MPI_COMM_WORLD : comm); 

  /* If this PE is inactive, return now */
  if (local_vec_length <= 0) return(env);

  /* Compute global length as sum of local lengths */
  n = local_vec_length;
  MPI_Allreduce(&n, &Nsum, 1, PVEC_INTEGER_MPI_TYPE, MPI_SUM, comm);
  env->global_vec_length = Nsum;

  /* Check input global length against computed value */
  if (Nsum != global_vec_length) {
    printf(BAD_N);
    PVecFreeMPI(env);
    return(NULL);
    } 

  /* Return the pointer env */
  return(env);
}


void PVecFreeMPI(void *machEnv)
{
  machEnvType env;

  env = (machEnvType) machEnv;
  if (env == NULL) return;

  if (!(env->init_by_user)) MPI_Finalize();

  free(machEnv);
}

 
N_Vector N_VNew(integer Ntotal, machEnvType machEnv)
{
  N_Vector v; 
  int N_local, N_global; 
  int i, j, ns;
  N_Vector *ptr; 

  if (Ntotal <= 0) return(NULL);
  if (machEnv == NULL) return(NULL);

  N_local = machEnv->local_vec_length;
  N_global = machEnv->global_vec_length;

  /* Determine the number of sensitivity vectors, ns */
  ns = (Ntotal / N_global) - 1;

  v = (N_Vector) malloc(sizeof *v); 
  if (v == NULL) return(NULL); 

  v->data = (real *) malloc((1+ns)*N_local*sizeof(real)); 
  if (v->data == NULL) { 
    free(v); 
    return(NULL); 
  } 

  v->length = N_local*(1+ns); 
  v->global_length = N_global*(1+ns); 
  v->machEnv = machEnv; 

  ptr = (N_Vector *) malloc((1+ns)*sizeof(N_Vector));
  if (ptr == NULL) {
    free(v->data);
    free(v);
    return(NULL);
  }
  for (i = 0; i <= ns; i++){ 
    ptr[i] = (N_Vector) malloc(sizeof(**ptr)); 
    if (ptr[i] == NULL){ 
      for (j = (i-1); j >= 0; j--) {
	free(ptr[j]);
      }
      free(ptr);
      free(v->data);
      free(v); 
      return(NULL); 
    }
    (ptr[i])->data = v->data + (i*N_local); 
    (ptr[i])->length = N_local; 
    (ptr[i])->global_length = N_global; 
    (ptr[i])->machEnv = machEnv; 
    (ptr[i])->subptr = NULL; 
  } 
  v->subptr = ptr;

  return(v);
}


void N_VFree(N_Vector x)
{
  int Ntotal, N_global;
  int i, ns;
  machEnvType machEnv;
  N_Vector *ptr;

  Ntotal = x->global_length;

  machEnv = (machEnvType) x->machEnv;
  N_global = machEnv->global_vec_length;

  /* Determine the number of sensitivity vectors, ns */
  ns = (Ntotal / N_global) - 1;

  ptr = (N_Vector *) x->subptr;
  for (i = 0; i <= ns; i++) {
    free(ptr[i]);
  }
  free(x->subptr); 
  free(x->data);         
  free(x);
  return;
}

