/******************************************************************
 *                                                                *
 * File          : sensitivity.h                                  *
 * Last Modified : 20 August 1999                                 *
 *----------------------------------------------------------------*
 * This is the header file for the implementation of computing    *
 * sensitivity information as given in the file:                  *
 * ../source/sensitivity.c                                        *
 *                                                                *
 ******************************************************************/


#ifdef __cplusplus     /* wrapper to enable C++ usage */
extern "C" {
#endif

#ifndef _sensitivity_h
#define _sensitivity_h

#include "cvode.h"
#include "llnltyps.h" 
#include "sens_cvspgmr.h"

/******************************************************************
 *                                                                *
 * Function : SensCVodeMalloc                                     *
 *----------------------------------------------------------------*
 * SensCVodeMalloc initializes a block of sensitivity data.       *
 * It also allocates and initializes memory for a problem         *
 * to be solved by CVODE.                                         *
 *                                                                *
 * Ntotal  is the total number of ODEs to be solved by CVODE.     *
 *                                                                *
 * f       is the right hand side function in y' = f(t,y).        *          
 *                                                                *
 * t0      is the initial value of t.                             *
 *                                                                *
 * y0      is the initial condition vector y(t0).                 *
 *                                                                *
 * lmm     is the type of linear multistep method to be used.     *
 *            The legal values are ADAMS and BDF (see previous    *
 *            description).                                       *
 *                                                                *
 * iter    is the type of iteration used to solve the nonlinear   *
 *            system that arises during each internal time step.  *
 *            The legal values are FUNCTIONAL and NEWTON.         *
 *                                                                *
 * itol    is the type of tolerances to be used.                  *
 *            The legal values are:                               *
 *               SS (scalar relative and absolute  tolerances),   *
 *               SV (scalar relative tolerance and vector         *
 *                   absolute tolerance).                         *
 *                                                                *
 * reltol  is a pointer to the relative tolerance scalar.         *
 *                                                                *
 * abstol  is a pointer to the absolute tolerance scalar or       *
 *            an N_Vector tolerance.                              *
 *                                                                *
 * f_data  is a pointer to user data that will be passed to the   *
 *             user's f function every time f is called.          *
 *                                                                *
 * errfp   is the file pointer for an error file where all CVODE  *
 *            warning and error messages will be written. This    *
 *            parameter can be stdout (standard output), stderr   *
 *            (standard error), a file pointer (corresponding to  *
 *            a user error file opened for writing) returned by   *
 *            fopen, or NULL. If the user passes NULL, then all   *
 *            messages will be written to standard output.        *
 *                                                                *
 * optIn   is a flag indicating whether there are any optional    *
 *            inputs from the user in the arrays iOpt and rOpt.   *
 *            Pass FALSE to indicate no optional inputs and TRUE  *
 *            to indicate that optional inputs are present.       *
 *                                                                *
 * iopt    is the user-allocated array (of size OPT_SIZE given    *
 *            later) that will hold optional integer inputs and   *
 *            outputs.  The user can pass NULL if he/she does not *
 *            wish to use optional integer inputs or outputs.     *
 *            If optIn is TRUE, the user should preset to 0 those *
 *            locations for which default values are to be used.  *
 *                                                                *
 * ropt    is the user-allocated array (of size OPT_SIZE given    *
 *            later) that will hold optional real inputs and      *
 *            outputs.  The user can pass NULL if he/she does not *
 *            wish to use optional real inputs or outputs.        *
 *            If optIn is TRUE, the user should preset to 0.0 the *
 *            locations for which default values are to be used.  *
 *                                                                *
 * machEnv is a pointer to machine environment-specific           *
 *            information.                                        *
 *                                                                *
 * ny      is the number of state variables in the ODE system.    *
 *                                                                *
 * ns      is the number of sensitivity vectors to be solved.     *
 *                                                                *
 * p       is the pointer to the array of real parameters.        *
 *                                                                *
 * pbar    is a pointer to an array of nonzero, real parameters   *
 *            that are used to scale the sensitivity ODEs.        *
 *                                                                *
 * rhomax  is a real value used for selecting a finite difference *
 *            formula to estimate the scaled sensitivity          *
 *            derivative w'_i, where 1 <= i <= ns.                *
 *                                                                *
 *----------------------------------------------------------------*
 *** Scaled sensitivity derivatives w' are estimated using finite *
 *** difference formulas involving the function f(t, y, p) and    *
 *** forward (or backward) perturbations, such as y + deltay*w    *
 *** and p + deltap*pbar. Vectors y and p can be perturbed        *
 *** separately, or simultaneously if the same value is used for  *
 *** deltay and deltap.                                           *
 ***                                                              *
 *** The heuristics for selecting deltay and deltap are based on  *
 *** the machine unit roundoff (uround), the user-specified       *
 *** relative error tolerance (rtol), and a weighted RMS norm of  *
 *** the scaled sensitivity vector (wrmsnorm(w)):                 *
 ***                                                              *
 ***   deltap = sqrt(MAX(rtol, uround),                           *
 ***                                                              *
 ***   deltay = 1 / (MAX(wrmsnorm(w), 1/deltap)).                 *
 ***                                                              *
 *** Additonally, this subroutine computes                        *
 ***                                                              *
 ***   ratio = MAX(deltay/deltap, deltap/deltay)                  *
 ***                                                              *
 *** to assess if deltay and deltap have comparable magnitudes.   *
 *----------------------------------------------------------------*
 *                                                                *
 *        rhomax = 0: Choose delta = MIN(deltay, deltap) and      *
 *               use 1 centered difference to estimate w'.        *
 *               Cost: 2 evaluations of f(t, y, p).               *
 *               Accuracy: 2nd order.                             *
 *                                                                *
 *        0 < rhomax < 1: Use centered differences on each        *
 *               term: Jw, pbar * df/dp.                          *
 *               Cost: 4 evaluations of f(t, y, p).               *
 *               Accuracy: 2nd order.                             *
 *                                                                *
 *        rhomax >= 1: If  (ratio > rhomax), use centered         *
 *               differences on each term: Jw, pbar * df/dp.      *
 *               Otherwise, choose delta = MIN(deltay, deltap)    *
 *               and use 1 centered difference to estimate w'.    *
 *               Cost: 2 or 4 evaluations of f(t, y, p).          *
 *               Accuracy: 2nd order.                             *
 *                                                                *
 *       -1 < rhomax < 0: Use forward differences on each         *
 *               term: Jw, pbar * df/dp.                          *
 *               Cost: 2 evaluations of f(t, y, p).               *
 *               Accuracy: 1st order.                             *
 *                                                                *
 *       rhomax <= -1: If  (ratio > abs(rhomax)), use forward     *
 *               differences on each term: Jw, pbar * df/dp.      *
 *               Otherwise, choose delta = MIN(deltay, deltap)    *
 *               and use 1 forward difference to estimate w'.     *
 *               Cost: 1 or 2 evaluations of f(t, y, p).          *
 *               Accuracy: 1st order accuracy.                    *
 *                                                                *
 *----------------------------------------------------------------*
 *                                                                *
 * Note: The tolerance values may be changed in between calls to  *
 *       CVode for the same problem. These values refer to        *
 *       (*reltol) and either (*abstol), for a scalar absolute    *
 *       tolerance, or the components of abstol, for a vector     *
 *       absolute tolerance.                                      *
 *                                                                * 
 * If successful, SensCVodeMalloc returns a pointer to the        *
 * initialized problem memory. This pointer should be passed to   *
 * CVode. If an initialization error occurs, SensCVodeMalloc      *
 * prints an error message to the file specified by errfp and     *
 * returns NULL.                                                  *
 *                                                                *
 ******************************************************************/


void *SensCVodeMalloc(integer Ntotal, RhsFn f, real t0, N_Vector y0, 
		      int lmm, int iter, int itol, real *reltol, 
		      void *abstol, void *f_data, FILE *errfp, 
		      boole optIn, long int iopt[], real ropt[], 
		      void *machEnv, int ny, int ns, void *p, 
		      void *pbar, real rhomax);


/******************************************************************
 *                                                                *
 * Function : SensCVodeFree                                       *
 *----------------------------------------------------------------*
 * SensCVodeFree frees the problem memory cvode_mem allocated by  *
 * SensCVodeMalloc. Its only argument is the pointer cvode_mem    *
 * returned by SensCVodeMalloc.                                   *
 *                                                                *
 ******************************************************************/

void SensCVodeFree(void *cvode_mem);


/***************************************************************
 *                                                             *
 * Function : SensSetAtol                                      *
 *-------------------------------------------------------------*
 * SensSetAtol initializes the absolute error tolerances for   *
 * the scaled sensitivity vectors to be the same as the        *
 * absolute error tolerances for the state vectors.            *
 *                                                             *
 * atol is the pointer to the array of absolute error          *
 *         tolerances.                                         *
 *                                                             *
 * ns   is the number of sensitivity vectors to be computed.   *
 *                                                             *
 ***************************************************************/

void SensSetAtol(N_Vector atol, integer ns);


/***************************************************************
 *                                                             *
 * Type : SensData                                             *
 *-------------------------------------------------------------*
 * The type SensData is a type for the block of data required  *
 * for computing the sensitivity of the ODE solution with      *
 * respect to its parameters. This block of data is created    *
 * by a user call to SensCVodeMalloc.                          *
 *                                                             *
 * cv_mem is the pointer to CVODE memory returned by           *
 *           SensCVodeMalloc.                                  *
 *                                                             *
 * f_ptr  is the pointer to the right hand side function in    *
 *           y' = f(t,y).                                      *
 *                                                             *
 * f_data is the pointer to user data that will be passed to   *
 *           the user's f function every time f is called.     *
 *                                                             *
 * p      is the pointer to an array of real parameters.       *
 *                                                             *
 * pbar   is a pointer to an array of nonzero, real parameters *
 *           that are used to scale the sensitivity ODEs.      *
 *                                                             *
 * rhomax is a real value used for selecting a finite          *
 *           difference formula to estimate the scaled         *
 *           sensitivity derivative w' = Jw + pbar * df/dp.    *
 *                                                             *
 * ny     is the number of state variables in the ODE system.  *
 *                                                             *
 * ns     is the number of sensitivity vectors to be solved.   *
 *                                                             *
 * ftemp  is temporary storage for estimating sensitivity      *
 *           derivatives using finite differences.             *
 *                                                             *
 * ytemp  is temporary storage for estimating sensitivity      *
 *           derivatives using finite differences.             *
 *                                                             *
 ***************************************************************/
 
typedef struct {
  void *cv_mem;
  void (*f_ptr)(integer, real, N_Vector, N_Vector, void *);
  void *f_data;
  real *p;
  real *pbar;
  real rhomax;
  integer ny;
  integer ns; 
  N_Vector ftemp;
  N_Vector ytemp;
  void *ad_data, *ad_u, *ad_udot;
} *SensData;

#endif

#ifdef __cplusplus
}
#endif
